"""
Interface functions for Sigma-UQ. They are trying to replicate
the commands of type uq_xxx in UQLab.
"""
import numpy as np
import adranis_sigma.uq.helpers as helpers



class uq():
    """The UQ interface class
    """

    def __init__(self, session):
        self.session = session

    def createModule(self, of_type, Opts, nargout=1):
        """Generic method for creating a UQ Object aka
        UQ module in UQLab's terminology.

        Args:
            of_type (string): The type of object to create
            Opts (dict): The creation options
            nargout (int, optional): number of output arguments
                that are expected from the equivalent uq_createX
                command that will be executed in the UQLab worker
                on the cloud. Defaults to 1.

        Returns:
           dict: A dictionary of the UQ Object that was created
        """
        REQ = {
            'Command': f"uq_create{of_type.lower().title()}",
            'Argument': [{
                'Value': Opts
            }],
            'nargout': nargout,
        }
        resp = self.session.rest_call(request=REQ)
        return self.session.get_resp_value(resp)
        # return True

    def createInput(self, InputOpts):
        self.session.logger.debug("Starting the processing of Input options.")
        if "Marginals" in InputOpts:
            for idx, marginal in enumerate(InputOpts["Marginals"]):
                PARAMS_SPECIFIED = "Parameters" in marginal and marginal[
                    "Parameters"]
                MOMENTS_SPECIFIED = "Moments" in marginal and marginal[
                    "Moments"]
                self.session.logger.debug(
                    f"Marginal nr. {idx}: Parameters specified: {PARAMS_SPECIFIED}, "
                        "Moments specified: {MOMENTS_SPECIFIED}")
                # Address the issue of mixed Moments-Parameters specification:
                # The fix here is to give the non-specified one as empty list
                if PARAMS_SPECIFIED and MOMENTS_SPECIFIED:
                    msg = "Cannot have both parameters and moments specified for a Marginal!"
                    raise RuntimeError(msg)
                if PARAMS_SPECIFIED:
                    # make sure that Parameters are like this: [[a,b]] and not
                    # like this: [a,b]
                    if not isinstance(marginal["Parameters"][0], list):
                        marginal["Parameters"] = [marginal["Parameters"]]
                    marginal["Moments"] = []
                if MOMENTS_SPECIFIED:
                    # make sure that Moments are like this: [[a,b]] and not
                    # like this: [a,b]
                    if not isinstance(marginal["Moments"][0], list):
                        marginal["Moments"] = [marginal["Moments"]]
                    marginal["Parameters"] = []
        return self.createModule('input', InputOpts)

    def createModel(self, ModelOpts):
        # check the type of model:
        # Model: exists locally UNLESS it is an mString
        # MetaModel: exists remotely
        if ModelOpts['Type'].lower() in ['uq_default_model', 'model'
                                         ] and 'mString' not in ModelOpts:
            # assign a compatible value to Type
            ModelOpts['Type'] = 'uq_default_model'
            # local definition
            ModelOpts['mFile'] = 'uq_cloud_remote'
            # In this case the 'ModelFun' needs to be supplied to carry out
            # model evaluations
            assert 'ModelFun' in ModelOpts, 'The ModelFun field is required and has not been supplied!'
            if 'Parameters' in ModelOpts:
                ModelOpts['Parameters']['theModel'] = ModelOpts['ModelFun']
            else:
                ModelOpts['Parameters'] = {'theModel': ModelOpts['ModelFun']}
            return self.createModule('model', ModelOpts)
        else:
            if ModelOpts['Type'].lower() in ['uq_default_model', 'model']:
                ModelOpts['Type'] = 'uq_default_model'
            # remote definition
            return self.createModule('model', ModelOpts)

    def createAnalysis(self, AnalOpts):
        return self.createModule('analysis', AnalOpts)

    def getSample(self, Input=None, N=None, Method='MC', nargout=1):

        REQ = {'Command': 'uq_getSample', 'nargout': nargout}
        if Input is None:
            REQ['Argument'] = [{'Value': N}, {'Value': Method}]
        else:
            if isinstance(Input, str):
                REQ['Argument'] = [{
                    'Value': Input
                }, {
                    'Value': N
                }, {
                    'Value': Method
                }]
            elif isinstance(Input, dict):
                REQ['Argument'] = [{
                    'Value': Input['Name']
                }, {
                    'Value': N
                }, {
                    'Value': Method
                }]
            else:
                raise ValueError(f"Unsupported type of Input: {str(type(Input))}!")
        resp = self.session.rest_call(request=REQ)
        resp_value = self.session.get_resp_value(resp)

        if nargout <= 1:
            theSamples = np.array(resp_value, ndmin=2)
            # In case theSamples is a row vector we have to make a consistency check:
            # Is this correct (i.e. M>1 and N=1) or should it be transposed? (M=1 N>1)
            # For this we need some info from Input
            if theSamples.shape[0] == 1 and nargout == 1:
                if Input is None:
                    # in this case we really need the input so we have to make one more call
                    # to sigma
                    Input = self.getInput('')

                if isinstance(Input['Marginals'], dict):
                    M = 1
                else:
                    M = sum(
                        isinstance(x.get('Type'), str) for x in Input['Marginals'])
                if M == 1:
                    return theSamples.T
                else:
                    return theSamples
            else:
                return theSamples
        else:
            for idx, respi in enumerate(resp_value):
                respi = np.array(respi, ndmin=2)
                if N > 1 and respi.shape[0] == 1:
                    resp_value[idx] = respi.T
                else:
                    resp_value[idx] = respi
            return resp_value

    def evalModel(self, Model=None, X=None, nargout=1):
        if X is None:
            return None
        # make sure that X is correctly interpreted when a single sample is
        # provided
        X = np.array(X, ndmin=2)
        if Model is None:
            # If no Model is given get the currently selected one within the UQ
            # session
            Model = self.getModel('')
        # Check the type of model
        isLOCAL = Model['Type'].lower() in ['model', 'uq_default_model']
        isLOCAL = isLOCAL and 'mString' not in Model
        if not isLOCAL:
            REQ = {
                'Command':
                'uq_evalModel',
                'Argument': [{
                    'Value': Model['Name']
                }, {
                    'Value': helpers.jsonify_np_array(X)
                }],
                'nargout':
                nargout
            }

            # if we reached this far the model exists on the uq_worker
            resp = self.session.rest_call(request=REQ)
            resp_value = self.session.get_resp_value(resp)
        else:
            # Check whether there are parameters involved
            PARAMS_EXIST = len(Model['Parameters']) > 1
            # Is the model vectorized?
            IS_VECTORIZED = Model['isVectorized']
            # Assume that 'ModelFun' field exists, but make sure it is callable
            if PARAMS_EXIST and IS_VECTORIZED:
                resp_value = helpers.function_eval(Model['ModelFun'], X,
                                                   Model['Parameters'])
            elif PARAMS_EXIST and not IS_VECTORIZED:
                resp_value = []
                for x in X:
                    resp_value.append(
                        helpers.function_eval(Model['ModelFun'], x,
                                              Model['Parameters']))
                resp_value = np.array(resp_value)
            elif not PARAMS_EXIST and IS_VECTORIZED:
                resp_value = helpers.function_eval(Model['ModelFun'], X)
            else:  # not PARAMS_EXIST and not IS_VECTORIZED
                resp_value = []
                for x in X:
                    resp_value.append(
                        helpers.function_eval(Model['ModelFun'], x))
                resp_value = np.array(resp_value)

        # we have to make a consistency check
        if nargout > 1:
            for idx, respi in enumerate(resp_value):
                respi = np.array(respi, ndmin=2)
                if X.shape[0] > 1 and respi.shape[0] == 1:
                    resp_value[idx] = respi.T
                else:
                    resp_value[idx] = respi
        if nargout == 1:
            resp_value = np.array(resp_value, ndmin=2)
            if X.shape[0] > 1 and resp_value.shape[0] == 1:
                resp_value = resp_value.T
        return resp_value

    def getModule(self, of_type, name):
        if of_type == 'input':
            cmd = 'uq_getInput'
        if of_type == 'model':
            cmd = 'uq_getModel'
        if of_type == 'analysis':
            cmd = 'uq_getAnalysis'
        REQ = {'Command': cmd, 'Argument': [{'Value': name}]}
        resp = self.session.rest_call(request=REQ)
        resp_value = self.session.get_resp_value(resp)
        return resp_value

    def getInput(self, name):
        return self.getModule('input', name)

    def getModel(self, name):
        return self.getModule('model', name)

    def getAnalysis(self, name):
        return self.getModule('analysis', name)

    def getAnalysisResults(self, name):
        REQ = {
            'Command': 'uq_getAnalysisResults',
            'Argument': [{
                'Value': name
            }]
        }
        resp = self.session.rest_call(request=REQ)
        resp_value = self.session.get_resp_value(resp)
        return resp_value

    def listModules(self, of_type):
        REQ = {
            'Command': 'uq_listModules',
            'Argument': [{
                'Value': of_type
            }],
            'nargout': 3
        }
        resp = self.session.rest_call(request=REQ)
        print(resp[2]['Value'])

    def listInputs(self):
        self.listModules('input')

    def listModels(self):
        self.listModules('model')

    def listAnalyses(self):
        self.listModules('analysis')

    def selectModule(self, of_type, name):
        if of_type == 'input':
            cmd = 'uq_selectInput'
        if of_type == 'model':
            cmd = 'uq_selectModel'
        if of_type == 'analysis':
            cmd = 'uq_selectAnalysis'

        REQ = {'Command': cmd, 'Argument': [{'Value': name}]}
        resp = self.session.rest_call(request=REQ)
        resp_value = self.session.get_resp_value(resp)
        return resp_value

    def selectInput(self, name):
        self.selectModule('input', name)

    def selectModel(self, name):
        self.selectModule('model', name)

    def selectAnalysis(self, name):
        self.selectModule('analysis', name)

    def removeModule(self, name, module_type):
        REQ = {
            'Command': 'uq_removeModule',
            'Argument': [{
                'Value': name
            }, {
                'Value': module_type.lower()
            }]
        }
        resp = self.session.rest_call(request=REQ)
        resp_value = self.session.get_resp_value(resp)
        return resp_value

    def print(self, obj=None, name=None, of_type=None):
        if obj is None:
            moduleToPrint = {"Name": name, "Class": of_type}
        else:
            moduleToPrint = {"Name": obj["Name"], "Class": obj["Class"]}
        REQ = {'Command': 'uq_print', 'Argument': [{'Value': moduleToPrint}]}
        resp = self.session.rest_call(request=REQ)
        resp_value = self.session.get_resp_value(resp)
        print(resp_value)

    def display(self, obj=None, name=None, of_type=None, **kwargs):
        if obj is None:
            obj = self.getModule(of_type=of_type, name=name)
        return helpers.display(obj, self, **kwargs)

    def rng(self, seed=None, generator=None):
        if seed is None and generator is None:
            REQ = {
                'Command': 'uq_rng',
            }
        elif generator is None:
            # No generator specified but the seed is not None
            REQ = {'Command': 'uq_rng', 'Argument': [{'Value': seed}]}
        elif seed is None:
            # No seed specified but the generator is not None: return Error
            raise ValueError(
                'The seed value is required when specifying the generator.')
        else:
            # Both seed and generator specified
            REQ = {
                'Command': 'uq_rng',
                'Argument': [{
                    'Value': seed
                }, {
                    'Value': generator
                }]
            }
        resp = self.session.rest_call(request=REQ)
        resp_value = self.session.get_resp_value(resp)
        return resp_value

    def all_f(self, function=None, X=None, Marginals=None):
        if X is None or Marginals is None:
            return None
        REQ = {
            'Command':
            function,
            'Argument': [{
                'Value': helpers.jsonify_np_array(X)
            }, {
                'Value': Marginals
            }],
            'nargout':
            1
        }
        resp = self.session.rest_call(request=REQ)
        resp_value = np.array(self.session.get_resp_value(resp), ndmin=2)
        return resp_value.reshape(X.shape)

    def all_cdf(self, X=None, Marginals=None):
        return self.all_f(function='uq_all_cdf', X=X, Marginals=Marginals)

    def all_pdf(self, X=None, Marginals=None):
        return self.all_f(function='uq_all_pdf', X=X, Marginals=Marginals)

    def all_invcdf(self, F=None, Marginals=None):
        return self.all_f(function='uq_all_invcdf', X=F, Marginals=Marginals)

    def GeneralIsopTransform(self, X, X_Marginals, X_Copula, Y_Marginals,
                             Y_Copula):
        REQ = {
            'Command':
            'uq_GeneralIsopTransform',
            'Argument': [{
                'Value': helpers.jsonify_np_array(X)
            }, {
                'Value': X_Marginals
            }, {
                'Value': X_Copula
            }, {
                'Value': Y_Marginals
            }, {
                'Value': Y_Copula
            }],
            'nargout':
            1
        }
        resp = self.session.rest_call(request=REQ)
        resp_value = np.array(self.session.get_resp_value(resp), ndmin=2)
        return resp_value.reshape(X.shape)

    def extractObj(self, parentName, objPath, objType):
        REQ = {
            'Command':
            'uq_extractObj',
            'Argument': [{
                'Value': parentName
            }, {
                'Value': objPath
            }, {
                'Value': objType
            }],
            'nargout':
            1
        }
        resp = self.session.rest_call(request=REQ)
        resp_value = self.session.get_resp_value(resp)
        return resp_value

    def extractFromInput(self, parentName, objPath):
        return self.extractObj(parentName=parentName,
                               objPath=objPath,
                               objType="input")

    def extractFromModel(self, parentName, objPath):
        return self.extractObj(parentName=parentName,
                               objPath=objPath,
                               objType="model")

    def extractFromAnalysis(self, parentName, objPath):
        return self.extractObj(parentName=parentName,
                               objPath=objPath,
                               objType="analysis")

    def eval_Kernel(self, X1, X2, theta, Options):
        theta = np.asarray(theta)
        REQ = {
            'Command':
            'uq_eval_Kernel',
            'Argument': [{
                'Value': helpers.jsonify_np_array(X1)
            }, {
                'Value': helpers.jsonify_np_array(X2)
            }, {
                'Value': helpers.jsonify_np_array(theta)
            }, {
                'Value': Options
            }],
            'nargout':
            1
        }
        resp = self.session.rest_call(request=REQ)
        resp_value = self.session.get_resp_value(resp)
        return np.asarray(resp_value)

    # Input module, Copula-specific functions
    def PairCopula(self, family, theta, rotation=None):
        REQ = {
            'Command': 'uq_PairCopula',
            'Argument': [{
                'Value': family
            }, {
                'Value': theta
            }],
            'nargout': 1
        }
        if rotation is not None:
            REQ['Argument'].append({'Value': rotation})

        resp = self.session.rest_call(request=REQ)
        return self.session.get_resp_value(resp)

    def PairCopulaOperation(self, copula, operation, *args):
        REQ = {
            'Command': operation,
            'Argument': [{
                'Value': copula
            }],
            'nargout': 1
        }
        for arg in args:
            REQ["Argument"].append({'Value': arg})

        resp = self.session.rest_call(request=REQ)
        return self.session.get_resp_value(resp)

    def PairCopulaKendallTau(self, copula):
        """_summary_

        Args:
            copula (_type_): _description_

        Returns:
            _type_: _description_
        """
        return self.PairCopulaOperation(copula, 'uq_PairCopulaKendallTau')

    def PairCopulaUpperTailDep(self, copula):
        """_summary_

        Args:
            copula (_type_): _description_

        Returns:
            _type_: _description_
        """
        return self.PairCopulaOperation(copula, 'uq_PairCopulaUpperTailDep')

    def PairCopulaLowerTailDep(self, copula):
        """_summary_

        Args:
            copula (_type_): _description_

        Returns:
            _type_: _description_
        """
        return self.PairCopulaOperation(copula, 'uq_PairCopulaLowerTailDep')

    def CopulaSummary(self, copula, *args):
        """_summary_

        Args:
            copula (_type_): _description_

        Returns:
            _type_: _description_
        """
        return self.PairCopulaOperation(copula, 'uq_CopulaSummary', *args)

    def test_block_independence(self,
                                X,
                                alpha,
                                stat='Kendall',
                                correction='auto',
                                verbose=0,
                                nargout=1):
        """_summary_

        Args:
            X (_type_): _description_
            alpha (_type_): _description_
            stat (str, optional): _description_. Defaults to 'Kendall'.
            correction (str, optional): _description_. Defaults to 'auto'.
            verbose (int, optional): _description_. Defaults to 0.
            nargout (int, optional): _description_. Defaults to 1.

        Returns:
            _type_: _description_
        """
        assert isinstance(
            X, list), "Requirement not satisfied: X should be a list!"
        REQ = {
            'Command':
            'uq_PairCopula',
            'Argument': [{
                'Value': X
            }, {
                'Value': alpha
            }, {
                'Value': stat
            }, {
                'Value': correction
            }, {
                'Value': verbose
            }],
            'nargout':
            nargout
        }
        resp = self.session.rest_call(request=REQ)
        return self.session.get_resp_value(resp)



import time
import sys
import uuid
import logging
import json
import ssl
import urllib.request
import numpy as np
import adranis_sigma.uq.helpers as helpers
import adranis_sigma.uq.interface as interface


class uq_session():

    def __init__(self, name=None, log_level=None, log_format=None):
        # Try to retrieve the Sigma API configuration from a file if available
        theConfig = helpers.sigma_load_config()
        self.stored_host = theConfig['host']
        self.stored_token = theConfig['token']
        self.compute_url = 'undefined'
        self.update_url = 'undefined'
        self.update_intermid_url = 'undefined'
        self.timeout = 1
        self.refresh_period = 0.5
        self.token = 'undefined'
        self.ssl_context = ''

        if name is None:
            self.name = uuid.uuid4().hex
        else:
            self.name = name
        # Set the logging output format
        if log_format is None:
            # Set the default log format if not specified
            log_format = ' %(name)s :: %(levelname)-8s :: %(message)s'
        # Initialize a logger
        logging.basicConfig(stream=sys.stdout, format=log_format)
        self.logger = logging.getLogger(__name__)
        # Set the logger log level
        self.set_log_level(log_level=log_level)

    def set_log_level(self, log_level=None):
        if log_level is not None:
            level = logging.getLevelName(log_level.upper())
            self.logger.setLevel(level)

    def rename(self, new_name):
        name_prev = self.name
        self.name = new_name
        return f"Session renamed from {name_prev} to {new_name}."

    def throw_worker_error(self, msg):
        self.logger.error('UQLab worker error: %s', msg)

    def get_resp_value(self, resp):
        # Build a response list if resp contains more than 1 'Value' keys
        if sum([1 for d in resp if 'Value' in d]) > 1:
            resp_final = []
            for respi in resp:
                if respi['Errors']:
                    self.throw_worker_error(respi['Message'])
                    return -1
                else:
                    resp_final.append(respi['Value'])
            return resp_final
        else:
            if resp['Errors']:
                self.throw_worker_error(resp['Message'])
                return -1
            else:
                return resp['Value']

    def rest_call(self, request=None, url=None, httpobj=False):
        if url is None:
            url = self.compute_url
        if request is None:
            raise ValueError('Empty request, aborting rest call.')
        return self.rest_poll(request=request, httpobj=httpobj)

    def rest_poll(self, request, httpobj=False):
        # send request
        resp = self.rest_call_cloud(url=self.compute_url,
                                    request=request,
                                    httpobj=True)
        if resp.getcode() == 201:
            resp_string = resp.read().decode('utf-8')
            resp_json = json.loads(resp_string)
            jobID = int(resp_json['jobID'])
        else:
            jobID = -2

        if jobID < 0:
            self.throw_worker_error(
                'UQ Engine error: the submitted request could not be completed.'
            )

        update_url = f"{self.update_url}{str(jobID)}"
        timeout = time.time() + self.timeout
        # wait until timeout
        while True:
            resp = self.rest_call_cloud(url=update_url,
                                        request='',
                                        httpobj=True)
            if resp.getcode() == 200:
                # return result
                return json.loads(resp.read().decode('utf-8'))
            if resp.getcode() == 201:
                # there is an intermediate step

                intermid_info_json = json.loads(resp.read().decode('utf-8'))
                intermid_fun = intermid_info_json['function']
                intermid_input = intermid_info_json['data']
                self.logger.info(
                    "Received intermediate compute request, function: %s.",
                    intermid_fun)
                self.logger.info('Carrying out local computation...')
                if 'parameters' in intermid_info_json:
                    intermid_parameters = intermid_info_json['parameters']
                    res_intermid = helpers.function_eval(
                        intermid_fun, intermid_input, intermid_parameters)
                else:
                    res_intermid = helpers.function_eval(
                        intermid_fun, intermid_input)
                self.logger.info('Local computation complete.')
                req_intermid = {
                    'data': np.array(res_intermid).tolist(),
                    'jobID': jobID
                }
                self.logger.info(
                    'Starting transmission of intermediate compute results (%s)...',
                    np.array(res_intermid).shape)
                resp = self.rest_call_cloud(url=self.update_intermid_url,
                                            request=req_intermid,
                                            httpobj=True)
                self.logger.info('Intermediate compute results sent.')
                timeout = time.time() + self.timeout
            if time.time() > timeout:
                return {'Errors': True, 'Message': 'Timeout reached'}
            time.sleep(self.refresh_period)

    def rest_call_cloud(self, url, request, httpobj=False):
        req = urllib.request.Request(url)
        req.add_header('Authorization', f"Token {self.token}")
        req.add_header('Content-Type', 'application/json; charset=utf-8')
        try:
            jsondataasbytes = request.encode('utf-8')
        except BaseException:
            jsondataasbytes = json.dumps(request,
                                         sort_keys=True).encode('utf-8')
        req.add_header('Content-Length', len(jsondataasbytes))
        response = urllib.request.urlopen(req,
                                          jsondataasbytes,
                                          context=self.ssl_context)
        if httpobj:
            return response
        else:
            return json.loads(response.read().decode('utf-8'))


class CloudSession(uq_session):

    def __init__(self,
                 host=None,
                 token=None,
                 name=None,
                 strict_ssl=True,
                 log_level="INFO",
                 log_format=None):
        super().__init__(name, log_level, log_format)
        if host is None:
            self.host = self.stored_host
        else:
            self.host = host
        if token is None:
            self.token = self.stored_token
        else:
            self.token = token
        self.refresh_period = 0.3  # in seconds
        self.timeout = 120  # in seconds
        self.compute_url = f"{self.host}/compute/"
        self.update_url = f"{self.host}/update/"
        self.update_intermid_url = f"{self.host}/update-intermid/"
        self.session_type = 'cloud'

        # In case the strict_ssl flag is set to False, requests to the Sigma
        # API will not involve checking the SSL certificate of the API. This can
        # be used to bypass certificate errors that may randomly occur in some
        # OS's
        if strict_ssl:
            self.ssl_context = None
        else:
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
            self.ssl_context = ctx
            # Raise a warning to remind the user that this is risky and they should
            # know what they are doing
            warn_msg = (
                "The SSL certificate of the API host will not be verified." +
                " Make sure that you understand the risks involved!")
            self.logger.warning(warn_msg)

        self.new()

    def new(self):

        self.cli = interface.uq(self)
        REQ = {
            'Command': 'disp',
            'Argument': [{
                'Value': 'Session started.'
            }],
            'nargout': 0,
        }
        resp = self.rest_call(REQ)
        if resp['Errors']:
            raise RuntimeError(resp['Message'])
        else:
            self.logger.info('A new session (%s) started.', self.name)

    def quit(self):
        REQ = {
            'Command': 'exit',
            'nargout': 0,
        }
        #resp = self.rest_call(REQ)
        resp = self.rest_call_cloud(url=self.compute_url,
                                    request=REQ,
                                    httpobj=False)
        if resp['Errors']:
            self.logger.error(
                'Something went wrong while terminating session %s.',
                self.name)
            return False
        else:
            self.logger.info('Session %s terminated.', self.name)
            self.name = None
            return True

    def list(self):
        REQ = {
            'Command': 'uq_list_sessions_web',
        }
        resp = self.rest_call(REQ)
        resp_value = self.get_resp_value(resp)
        print(resp_value)

    def reset(self):
        REQ = {'Command': 'uqlab', 'nargout': 0}
        resp = self.rest_call(REQ)
        if resp:
            self.logger.info('Reset successful.')
        else:
            self.logger.error('Reset failed.')

    def remove(self, name):
        REQ = {
            'Command': 'uq_delete_session_web',
            'Argument': [{
                'Value': name
            }],
            'nargout': 2
        }
        resp = self.rest_call(REQ)
        resp_value = self.get_resp_value(resp)
        if resp_value:
            if name == '*':
                self.logger.info('All sessions removed.')
            else:
                self.logger.info("Session %s  removed.", name)
        else:
            self.logger.error("Session %s  removal failed.", name)

    def save(self, name=None):
        if name is None:
            session_name = self.name
        else:
            session_name = name

        REQ = {
            'Command': 'uq_save_session_web',
            'Argument': [{
                'Value': session_name
            }]
        }
        resp = self.rest_call(REQ)
        resp_value = self.get_resp_value(resp)
        if resp_value:
            self.logger.info('Session %s saved.', session_name)
        else:
            self.logger.error('Tried to save Session %s but failed.',
                              session_name)

    def load(self, name):
        REQ = {'Command': 'uq_load_session_web', 'Argument': [{'Value': name}]}
        resp = self.rest_call(REQ)
        resp_value = self.get_resp_value(resp)
        self.name = name
        if resp_value:
            self.logger.info('Session %s loaded.', name)
        else:
            self.logger.error('Session %s loading failed.', name)

    def save_config(self):
        self.logger.info(helpers.sigma_save_config(self.host, self.token))



""" Helper functions for python-sigma-uq
"""
import os
from pathlib import Path
from importlib import import_module, util
import json
from appdirs import user_data_dir
import numpy as np
import adranis_sigma.uq.display_util as display_util
# Helper functions


def sigma_save_config(host, token):
    """ Saves the Sigma API url and token of the user
    to a file in a proper AppDir.
    If this specific file is found during import
    of the Sigma package, those values will be
    automatically loaded.

    Args:
        host (string): The URL of the SIGMA API endpoint
            that the user is authorized to connect to
        token (string): The token associated to the user

    Returns:
        string: (print friendly) Message related to the success of the
            operation
    """
    theDir = user_data_dir("Adranis-Sigma", "Sigma")
    # Create the folder if it doesn't exist
    Path(theDir).mkdir(parents=True, exist_ok=True)
    theFile = os.path.join(theDir, 'sigma_config.json')
    theConfig = {"host": host, "token": token}
    with open(theFile, 'w', encoding='UTF-8') as outfile:
        json.dump(theConfig, outfile)
    return f"Stored Sigma API configuration in {theFile}"


def sigma_load_config(verbose=False):
    # Loads the Sigma API url and token from the configuration file
    theDir = user_data_dir("Adranis-Sigma", "Sigma")
    theFile = os.path.join(theDir, 'sigma_config.json')
    if os.path.isfile(theFile):
        with open(theFile, 'r', encoding='UTF-8') as infile:
            theConfig = json.load(infile)
    else:
        theConfig = {"host": None, "token": None}
    if verbose:
        print(f"Loaded Sigma API configuration from file {theFile}.")
    return theConfig


def load_module(path, name):
    # Loads a module from a given path with a given name
    spec = util.spec_from_file_location(name, path)
    module = util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def jsonify_np_array(A):
    return A.tolist()


def getlines(fd):
    line = bytearray()
    c = None
    while True:
        c = fd.read(1)
        if c is None:
            return
        line += c
        if c == b'\n':
            yield str(line)
            del line[:]


def extract_json(input_string):
    # extracts the json raw sub-string from a string
    idx_start_candidates = [input_string.find('{'), input_string.find('[')]
    json_start = min(x for x in idx_start_candidates if x != -1)
    idx_end_candidates = [input_string.rfind('}'), input_string.rfind(']')]
    json_end = max(idx_end_candidates) + 1
    return input_string[json_start:json_end]


def function_eval(fun, X, Parameters=None):
    if callable(fun):
        return fun(X)
    else:
        # expecting the fun to be a string
        # we are expecting a syntax a.b.c (or longer) where a.b is the module and
        # c is the method
        X = np.array(X)
        method_sep = fun.rfind('.')
        the_module = import_module(fun[0:method_sep])
        the_method = getattr(the_module, fun[method_sep + 1:])
        if Parameters is None:
            return the_method(X)
        else:
            return the_method(X, Parameters)


def display(obj, theInterface, **kwargs):
    """Produces visualizations of UQ objects. Based on
    UQLab's uq_display.

    Args:
        obj (dict): a dictionary
            that is returned by Sigma API when a UQ object
            is created
        theInterface (CloudSession.interface): the interface
             of the cloud session

    Raises:
        ValueError: Unknown or Unsupported type of UQ object

    Returns:
        plotly figure: A handle to a plotly figure (e.g. fig)
        that can be visualized by `fig.show()`
    """
    if obj['Class'] == 'uq_input':
        # display input
        return display_util.Input(obj, theInterface, **kwargs)
    if obj['Class'] == 'uq_model':
        # display model
        if 'PCE' in obj:
            return display_util.PCE(obj)
        if 'Kriging' in obj:
            return display_util.Kriging(obj, theInterface)
        else:
            raise ValueError("Not yet implemented!")

    if obj['Class'] == 'uq_analysis':
        # display analysis
        print("Not yet implemented!")



""" Computational models that are used during unit testing
"""
import numpy as np


def XsinX(X):
    X = np.array(X, ndmin=2)
    return X * np.sin(X)


def hat2d(X):
    # 2D hat function
    # see https://www.uqlab.com/reliability-2d-hat-function
    X = np.array(X, ndmin=2)
    t1 = np.square(X[:, 0] - X[:, 1])
    t2 = 8 * np.power(X[:, 0] + X[:, 1] - 4, 3)
    return 20 - t1 - t2



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['adranis_sigma', 'adranis_sigma.uq']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'numpy>=1.22.4,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'plotly>=5.8.2,<6.0.0',
 'scipy>=1.8.1,<2.0.0']

setup_kwargs = {
    'name': 'adranis-sigma',
    'version': '0.1.0',
    'description': 'Firt release of the python binding of the Adranis Sigma API',
    'long_description': None,
    'author': 'Adranis GmbH',
    'author_email': 'contact@adranis.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



""" Plotting utility imitating the uq_display functionality of UQLab """
import numpy as np  # type: ignore
from scipy.stats import norm  # type: ignore
import pandas as pd  # type: ignore
import plotly.graph_objects as go  # type: ignore
import plotly.express as px  # type: ignore
import plotly.figure_factory as ff  # type: ignore
from plotly.subplots import make_subplots  # type: ignore
import plotly.io as pio  # type: ignore

# Figure template
pio.templates.default = 'simple_white'

# Template modifications
# Size
pio.templates[pio.templates.default].layout.height = 600
pio.templates[pio.templates.default].layout.width = 600

# Grid
pio.templates[pio.templates.default].layout.xaxis.showgrid = True
pio.templates[pio.templates.default].layout.yaxis.showgrid = True
pio.templates[pio.templates.default].layout.autosize = False

# Font
pio.templates[pio.templates.default].layout.font.size = 16

# Markers
pio.templates[pio.templates.default].data.histogram[0].marker.line.width = 0


def Input(Module,
          theInterface,
          inputs=None,
          plot_density=False,
          show_vine=False):
    '''Input display is a utility function for visualising a
    random vector specified in an Input object of generated
    by the uq_default_input module. By default, it generates
    a scatter plot matrix of the elements of the random vector.
    There is also the option to plot the PDF and CDF of the
    marginal distributions of the vector components.

    Args:
        Module (dict, optional): Input parameters of multivariate
            random vectors. Defaults to None.
        inputs (list, optional): Set of indexes that define
            the variables to be displayed . Defaults to None.
        plot_density (bool, optional): Flag to plot the
            marginal distributions. Defaults to False.
    '''

    # Parameters
    M_max = 10  # Input dimension limit
    N = 1000  # Sample size
    # Number of std's for the x_range in case of input_array
    # scalar (displaying of a single marginal)
    n_std = 5

    # Define inputs
    if isinstance(Module['Marginals'], dict):
        Module['Marginals'] = [Module['Marginals']]
    M = len(Module['Marginals'])
    if inputs is None:
        inputs = range(1, M + 1)
    inputs = [f'X{var}' for var in inputs]
    n_var = len(inputs)

    # Consistency checks
    if M > M_max:
        return print(
            'Requested input range is too large, please select a smaller range.'
        )

    # Get sample for plotting
    X = theInterface.getSample(Module, N, 'LHS')
    df = pd.DataFrame(X)
    df.columns = [f'X{var}' for var in range(1, X.shape[1] + 1)]
    df = df[inputs]

    # Produce the plot
    # Scatter plot for two inputs
    if (n_var == 2) & (not plot_density):
        # Figure
        fig = px.scatter(df,
                         x=df.columns[0],
                         y=df.columns[1],
                         marginal_x='histogram',
                         marginal_y='histogram',
                         opacity=0.4)
        # Aesthetics
        marker_color_ref = pio.templates[
            pio.templates.default].layout.colorway[1]
        fig._data_objs[1].marker.color = marker_color_ref
        fig._data_objs[2].marker.color = marker_color_ref
        fig.data[1]['opacity'] = fig.data[2]['opacity'] = 1

    # Scatter matrix for >2 inputs
    elif (n_var > 2) & (not plot_density):
        width = 250 * n_var if n_var <= 5 else 250 * 6
        # Figure
        fig = ff.create_scatterplotmatrix(df,
                                          diag='histogram',
                                          marker_opacity=0.4,
                                          width=width,
                                          height=width)
        # Aesthetics
        fig.layout.hovermode = False
        # ..Scatter color
        fig.layout.colorway = [
            pio.templates[pio.templates.default].layout.colorway[0]
        ]
        # ..Histogram color
        for data in fig.data:
            if type(data) == type(go.Histogram()):
                data.marker.color = pio.templates[
                    pio.templates.default].layout.colorway[1]

    # Line plots for density functions
    else:
        # Figure
        fig = make_subplots(rows=n_var,
                            cols=2,
                            vertical_spacing=0.12,
                            subplot_titles=('PDF', 'CDF'))
        row = 1
        for var in inputs:
            # Prepare data
            marginal = [x for x in Module['Marginals'] if x['Name'] == var][0]
            x_mean, x_std = marginal['Moments']
            x = np.linspace(x_mean - n_std * x_std, x_mean + n_std * x_std, N)

            # Calculate PDF and CDF
            pdf = theInterface.all_pdf(x, marginal)
            cdf = theInterface.all_cdf(x, marginal)

            # Add data to figure
            fig.add_trace(go.Scatter(
                x=x,
                y=pdf,
                line=dict(color=pio.templates[
                    pio.templates.default].layout.colorway[row - 1]),
                name=f'{var} PDF'),
                          row=row,
                          col=1)

            fig.add_trace(go.Scatter(
                x=x,
                y=cdf,
                line=dict(color=pio.templates[
                    pio.templates.default].layout.colorway[row - 1]),
                name=f'{var} CDF'),
                          row=row,
                          col=2)
            fig.update_xaxes(title_text=var, row=row, col=1)
            fig.update_xaxes(title_text=var, row=row, col=2)
            fig.update_yaxes(title_text=f'f<sub>{var}', row=row, col=1)
            fig.update_yaxes(title_text=f'F<sub>{var}', row=row, col=2)
            row += 1

        # Aesthetics
        fig.update_layout(showlegend=False, height=500 * n_var, width=1000)

    return fig


def PCE(Model):

    # Prepare data
    Coefficients = np.array(Model['PCE']['Coefficients'])
    nnz_idx = np.abs(Coefficients) > 0
    Coefficients = Coefficients[nnz_idx]
    Indices = np.array(Model['PCE']['Basis']['Indices'])
    Indices = Indices[nnz_idx]
    Indices_row_num = np.arange(1, Indices.shape[0] + 1)
    Degree = Indices.sum(axis=1)
    legend_txt = ['Mean', '<i>p=1', '<i>p=2', '<i>p=3', '<i>p>3']
    logCoeffs = np.log10(np.abs(Coefficients))
    logCoeffRange = np.max([1, np.max(logCoeffs) - np.min(logCoeffs)])

    # Create figure traces
    traces = []
    for k in range(5):
        if k < 4:
            idx = Degree == k
        else:
            idx = Degree >= k
        traces.append(
            go.Scatter(x=Indices_row_num[idx],
                       y=np.log10(np.abs(Coefficients[idx])),
                       mode='markers',
                       marker=dict(size=15 - 2 * k, opacity=0.5),
                       name=legend_txt[k]))

    # Define figure layout
    layout = go.Layout(title=f'NNZ Coeffs: {Coefficients.shape[0]}',
                       xaxis_title='𝛼',
                       yaxis_title='log<sub>10</sub>(|𝑦<sub>𝛼</sub>|)')
    # Create figure
    fig = go.Figure(data=traces, layout=layout)

    return fig


def Kriging(Model, theInterface):
    """Display Kriging dictionary with 1-dimensional output"""

    N1d = 500  # Number of points to plot in one-dimension
    N2d = 80  # Number of points to plot in two-dimension

    # One-dimensional case

    # Compute points to plot
    X = np.array(Model['ExpDesign']['X'])
    Y = np.array(Model['ExpDesign']['Y'])
    Xmin = np.min(X)
    Xmax = np.max(X)
    Xval = np.linspace(Xmin, Xmax, N1d).transpose()
    # Exp. design points belong to evaluation
    Xval = np.sort(np.concatenate((Xval, X)))
    [Ymu_krg,
     Ysigma_krg] = theInterface.evalModel(Model,
                                          Xval.reshape([Xval.shape[0], 1]),
                                          nargout=2)

    # Compute upper and lower bounds of the confidence interval
    conf_level = 0.95  # 95% confidendence level
    conf_interval = norm.ppf(1 - 0.5 *
                             (1 - conf_level), 0, 1) * np.sqrt(Ysigma_krg)

    # Figure
    # Traces
    traces = [
        go.Scatter(x=Xval.squeeze(),
                   y=Ymu_krg.squeeze(),
                   name='Kriging approximation',
                   fill=None,
                   mode='lines',
                   legendgroup=0),
        go.Scatter(x=Xval.squeeze(),
                   y=(Ymu_krg - conf_interval).squeeze(),
                   name='95% confidence interval',
                   fill=None,
                   legendgroup=1,
                   mode='lines',
                   marker=dict(color='lightgrey'),
                   showlegend=False),
        go.Scatter(x=Xval.squeeze(),
                   y=(Ymu_krg + conf_interval).squeeze(),
                   fill='tonexty',
                   name='95% confidence interval',
                   legendgroup=1,
                   mode='lines',
                   marker=dict(color='lightgrey')),
        go.Scatter(x=X.squeeze(),
                   y=Y.squeeze(),
                   name='Observations',
                   fill=None,
                   legendgroup=2,
                   mode='markers',
                   marker=dict(color='black'))
    ]

    # Layout
    layout = go.Layout(xaxis_title='𝑋<sub>1',
                       yaxis_title='𝑌',
                       hovermode='x',
                       width=800)

    # Create figure
    fig = go.Figure(data=traces, layout=layout)

    return fig
